/*
--------------------------------------------------------------------------------
Title:              AmpEd_Arduino.ino
Description:        Final Arduino code meant to run on the Arduino with the
                    AmpEd platform.
Authors:            Taylor Yee, Daniel Beckett
Created:            04/07/2020
Last Modified:      04/12/2020
Helpful Link(s):    https://forum.arduino.cc/index.php?topic=447906.0
                    http://henrysbench.capnfatz.com/henrys-bench/arduino-projects-tips-and-more/arduino-serial-basics-waiting-to-receive-a-serial-command/
                    https://forum.arduino.cc/index.php?topic=39965.0
--------------------------------------------------------------------------------
*/

/****** Processing Declarations ******/


// String representation of data collected - used when session ends
String results = "";

// Character variables to hold "command signals" from Processing
char presetInput;
char doneRecording;


/****** Control Scheme Declarations ******/
/*library for sabertooth: Arduino Libraries for SyRen/Sabertooth Serial, https://www.dimensionengineering.com/info/arduino */
#include <Sabertooth.h>      

Sabertooth ST(128);          //Initialize Sabertooth address

//Declaring pins
const int Y_pin = A0;        //pin A0: Analog input, Joystick x-axis 
const int X_pin = A1;        //pin A1: Analog input, Joystick y-axis

//HEADSWITCHS
const int RHS = 2;           //pin D2: Digital input, Right Head Switch
const int LHS = 4;           //pin D4: Digital input, Left  Head Switch

//Variables for holding analog pin inputs from joystick
int Y_axis;
int X_axis;
int Radius;                  //Distance of joystick from center

//Variables for ramping turn speeds
int Ramp_left;               //Counter for gradually increasing turn speed
int Ramp_right;              //Counter for gradually increasing turn speed

//Variables for holding mapped PWM outputs for motor driver
int Drive_power;
int Turn_power;

//Function definitions
void regJoystickCtrlScheme();
void headSwitchCtrlScheme();
void uniJoystickCtrlScheme();

/****** Accelerometer Declarations ******/
/*
#include <Wire.h>             
#include <SPI.h>
#include <SparkFunLSM9DS1.h>       //$$$$$Sketch-> Include Library-> Manage Libraries-> LSM9DS1

LSM9DS1 imu;                       //Create object: imu

// Using I2C communication, therefore our addresses are:
#define LSM9DS1_M  0x1E
#define LSM9DS1_AG 0x6B

//Settings for Arduino IDE Serial Monitor
#define PRINT_CALCULATED           //Comment this line to remove accelerometer data from serial monitor
#define PRINT_SPEED 250            //250 ms between prints
//#define DECLINATION -10.16       //Declination (degrees) in Flagstaff on 4/11/2020.  //likely useless

//Pin declaration
const int INT1_PIN_THS = 6;        //INT1 sends interrupt flag to D6 on Uno
*/
//Variables to hold data needed
int accelCollisions;               //$$$$ count coming from accelerometer $$$$

//Function definitions
void ACCELCollisions();
void printAccel();
void printStats();

//DEFINITIONS USED FOR TESTING
static unsigned long lastPrint = 0; //Keep track of print time

/****** IR Sensor Declarations *******/

/*SHARP GP2Y0A21YK0F IR sensor library https://www.makerguides.com/sharp-gp2y0a21yk0f-ir-distance-sensor-arduino-tutorial/ */
#include <SharpIR.h>

//Define model and input pin:
#define IR0Pin A2                  //front (sitting on chair facing motors)
#define IR1Pin A3                  //back
#define IR2Pin A4                  //left
#define IR3Pin A5                  //right
#define model 1080

// Create variable to store the distance (in cm):
int frontCm=0, backCm=0, leftCm=0, rightCm=0;
int distArray[] = {frontCm, backCm, leftCm, rightCm};
int ArrayValue=0, irCollisions=0;  //value in array, hit count
int i, minDist=10;                 //the min cm the sensors can read

//Create a new instance of the SharpIR class:
SharpIR frontSensor = SharpIR(IR0Pin, model);
SharpIR backSensor  = SharpIR(IR1Pin, model);
SharpIR leftSensor  = SharpIR(IR2Pin, model);
SharpIR rightSensor = SharpIR(IR3Pin, model);

// General variables
int totCollisions;                  //$$$$ total collisions coming from Accelerometer & IR$$$$
void IRCollisions();


/*******  DECLARATIONS    DECLARATIONS    DECLARATIONS    DECLARATIONS   DECLARATIONS  
                    SETUP   SETUP   SETUP   SETUP   SETUP   SETUP   SETUP   SETUP      ********/

void setup()
{
/****** Controls Setup******/
  SabertoothTXPinSerial.begin(9600);
  ST.autobaud();
  ST.drive(0);
  ST.turn(0);
  Drive_power = 0;
  Turn_power = 0;
  
/****** Head Switches ******/
  pinMode(RHS, INPUT);              //Pin 2 is the input for the Right Head Switch
  digitalWrite(RHS, HIGH);          //Set pull up resistor for pin 2
  
  pinMode(LHS, INPUT);              //Pin 4 is the input for the Left Head Switch
  digitalWrite(LHS, HIGH);          //Set pull up resistor for pin 4

/****** General Setup ******/

  Serial.begin(9600);               //I want to increase the BAUD rate eventually, maybe 115200  

/****** User Preference Setup ******/
  // from the Processing side
  while(!Serial.available())        //This loop means that we won't exit setup() until a preset has been sent
  {}
  
  // This will save the preset for use in loop()
//  if(Serial.available() == 'J' || Serial.available() == 'H')
//  {
    presetInput = Serial.read();
//  }
  // I know this looks like redundant code, but the above if-statement doesn't
  // seem to work without it, so leave this here
//  else
//  {
//    presetInput = Serial.read();
//  }

/****** Accelerometer Setup ******/
/*
  setupAccel();                       //Call function to setup the Accelerometer 

  if (imu.begin() == false)           //Check if Accelerometer is there
  {
    Serial.println("Failed to communicate with LSM9DS1.");
    Serial.println("Double-check wiring.");
    Serial.println("It's small and red");
    while (1);
  }
  
  // After turning the IMU on, configure the interrupts:
  configureLSM9DS1Interrupts();

*/  
}


/******* SETUP   SETUP  SETUP  SETUP   SETUP   SETUP   SETUP   SETUP  
     MAIN LOOP   MAIN LOOP   MAIN LOOP   MAIN LOOP   MAIN LOOP   MAIN LOOP *******/

void loop()
{
//  presetInput = 'H'; //testing

  chairOp(presetInput);         // This function call takes care of movement

  //ACCELCollisions();            // $$$$This function call updates accelerometer results (# of collisions)
  IRCollisions();
/**
 * Function called to write data to serial port if session has ended.
 * <p>
 * Note: Former "writeToGUI" method - this NEEDS to stay here
 * @param numCollisions number of times sensors detect "collision"
 * @param closestDistance closest distance (in cm?) WITHOUT colliding
 */
  if(Serial.available() > 0)
  {
    String results = "";
    doneRecording = Serial.read();

    if(doneRecording == 'S')
    {
      totCollisions = 0;
//      closestDistance = 0;
      
//      IRCollisions();         //$$$$$ start calculating collisions when starting to record $$$$$
//      ACCELCollisions();      //$$$$$ wouldnt we want these here????not outside the if statement$$$$$
    }

    if(doneRecording == 'D')    // Command signal for "done recording"
    {
      
      /**
       * Concatenates results to string, separated by comma
       * to make it readable when data reaches Processing end
       */
      totCollisions = irCollisions + accelCollisions;
      results = "Number of collisions: " + String(totCollisions);
      Serial.println(results);      // This sends info to Processing
    }
    
    else if(doneRecording == 'X')   // This allows user to go back and change the preset mid-session if they want
    {
      setup();
    }
  }
}

/****** MAIN LOOP   MAIN LOOP   MAIN LOOP   MAIN LOOP   MAIN LOOP   MAIN LOOP 
     FUNTIONS   FUNTIONS   FUNTIONS   FUNTIONS   FUNTIONS   FUNTIONS   FUNTIONS *****/

/*****Deciding Control Scheme ******/


void chairOp(char presetInput)
{

  // Movement code if user decided on regular joystick
  if(presetInput == 'J')
  {
    regJoystickCtrlScheme();
  }

  // Movement code if user decided on head switches
  else if(presetInput == 'H')
  {
    headSwitchCtrlScheme();
  }
}


/***** Regular Joystick Main Code *****/

void regJoystickCtrlScheme()
{
  //Read joystick positions
  Y_axis = analogRead(Y_pin); // Read Joysticks Y-axis analog value
  X_axis = analogRead(X_pin); // Read Joysticks X-axis analog value
  
  if (Y_axis > 550) {
    Drive_power = map(Y_axis, 1023, 550 , -127, -30);   //Reverse
  }
  
  else if (Y_axis < 470)  {
    Drive_power = map(Y_axis, 470, 0 , 30, 127);        //Forward
  }
  else {
    Drive_power = 0;                                    //Not driving
  }
                    
  if (X_axis < 470) {
    Turn_power = map(X_axis, 470, 0 , -30, -127);       //Turning left
  }
  else if (X_axis > 550) {
    Turn_power = map(X_axis, 550, 1023 , 30, 127);      //turning right
  }
  else {
    Turn_power=0;                                       //Not turning
  }
     
  //Set motor power outputs
  ST.drive(Drive_power); 
  ST.turn(Turn_power);
  delay(20);
}

/***** Head Switch Main Code ******/

void headSwitchCtrlScheme()
{
  uniJoystickCtrlScheme();
  
  //turning left
  if (!digitalRead(LHS)) {
    Turn_power = -65;
  }
  //turning right
  else if (!digitalRead(RHS)) {
    Turn_power = 65;
  }
  else {
    Turn_power = 0;
  }
  ST.turn(Turn_power);
  delay(20);
}

/****** Unidirectional Joystick Main Code *******/

void uniJoystickCtrlScheme()
{
  /*Radius = sqrt(abs(sq(512 - analogRead(X_pin)) - sq(512 - analogRead(Y_pin))));
  Drive_power = map(Radius, 10, 512, 20, 127);
  Drive_power = map(Radius, 0, 9, 0, 0);
  ST.drive(Drive_power);
  delay(20);

      /*  //TESTING
        // onboard LED blinks at 2sec intervals
        digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
        delay(2000);                       // wait for a second
        digitalWrite(LED_BUILTIN, LOW);    // turn the LED off by making the voltage LOW
        delay(2000);                       // wait for a second
*/
  //Forward caused by any value of joystick
  Y_axis = analogRead(Y_pin); // Read Joysticks Y-axis analog value
  X_axis = analogRead(X_pin); // Read Joysticks X-axis analog value


  //pulling back
  if (Y_axis > 560) {
    if (X_axis > Y_axis) { //is X more right than Y is back?
      Drive_power = map(X_axis, 1023, 560 , 127, 30);
    }
    else if ((1023 - X_axis) > Y_axis) { //is X more left than Y is back?
      Drive_power = map(X_axis, 460, 0 , 30, 127);
    }
    else {    //Y is furthest from center
      Drive_power = map(Y_axis, 1023, 560 , 127, 30);
    }
  }

  //Pushing foward
  else if (Y_axis < 460)  {
    if (X_axis < Y_axis) { //is X more left than Y is forward?
      Drive_power = map(X_axis, 460, 0 , 30, 127);
    }
    else if ((1023 - X_axis) < Y_axis) { //is X more right than Y is forward?
      Drive_power = map(X_axis, 1023, 560 , 127, 30);
    }
    else {    //Y is furthest from center
      Drive_power = map(Y_axis, 460, 0 , 30, 127);
    }
  }

  else {
    if (X_axis < 460) { //Y idle, X left
      Drive_power = map(X_axis, 460, 0 , 30, 127);
    }
    else if (X_axis > 560) { //Y idle, X right
      Drive_power = map(X_axis, 1023, 560 , 127, 30);
    }
    else {    //True idle
      Drive_power = 0;
    }
  }
  ST.drive(Drive_power);
}


/****** LSM9DS1 Main Code *******/
/*
void ACCELCollisions() //IN PROGRESS
{
  if (digitalRead(INT1_PIN_THS) == LOW)    //If the interrupt flag has been thrown
  {
    //printAccel();                          //Print "A: ax, ay, az" testing
    accelCollisions++;                     //Increase number of collisions
    Serial.println(accelCollisions);       //print that number

    /*
     * This next line fixed my code, 
     * but I'm not entirely sure how or if it's the best choice 
     * Consider: imu.readAccel();
     * Which would update the acceleration numbers, resetting the flag, hopefully?
     * Either way, we can't have one crash being registered as multiple.
     
    imu.getAccelIntSrc();                  //Update interrupt flag? This line fi
    //digitalWrite(INT1_PIN_THS, HIGH);

    //FOR TESTING
    //digitalWrite(LED_BUILTIN, HIGH);
  }

  //Prints out info, after interval "PRINT_SPEED=250"
  if ((lastPrint + PRINT_SPEED) < millis())
  {
    imu.readAccel();                       //Update acceleration values: axm ay, az
    //printAccel();                          //Print "A: ax, ay, az"
    Serial.println();
    lastPrint = millis();                  //Update lastPrint time
  }
  
}


/******* LSM9DS1 Interface Setup Function *********

uint16_t configureIMU()
{
  imu.settings.device.commInterface = IMU_MODE_I2C; // Set mode to I2C
  imu.settings.device.mAddress = LSM9DS1_M;         // Set mag address to 0x1E
  imu.settings.device.agAddress = LSM9DS1_AG;       // Set ag address to 0x6B
  imu.settings.gyro.enabled = false;                // Disable the gyro
  imu.settings.mag.enabled = false;                 // Disable magnetometer
  imu.settings.temp.enabled = false;                // Disable temperature sensor
  
  
  // Call imu.begin() to initialize the sensor and instill it with our new settings 
  //return imu.begin(LSM9DS1_AG_ADDR(1), LSM9DS1_M_ADDR(1), Wire); // set addresses and wire port
  return imu.begin(LSM9DS1_AG_ADDR(1), LSM9DS1_M_ADDR(1), Wire);
}

*/
/******** Accelerometer Setup Function *********/
/*
void setupAccel()
{
  Wire.begin();
  pinMode(INT1_PIN_THS,INPUT_PULLUP);               // input w/ pull up Resistor b/c active low interrupt
  
  int accelCollisions = 0;                          // $$$$ Collision count from accelerometer $$$$
  uint16_t status = configureIMU();                 //Configure accelerometer

  // [enabled] turns the acclerometer on or off.
  imu.settings.accel.enabled = true;                // Enable accelerometer
  // [enableX], [enableY], and [enableZ] can turn on or off
  // select axes of the acclerometer.
  imu.settings.accel.enableX = true;                // Enable X
  imu.settings.accel.enableY = true;                // Enable Y
  imu.settings.accel.enableZ = false;               // Disable Z
  // [scale] sets the full-scale range of the accelerometer.
  // accel scale can be 2, 4, 8, or 16
  imu.settings.accel.scale = 2; // Set accel scale to +/-2g.
  // [sampleRate] sets the output data rate (ODR) of the
  // accelerometer. ONLY APPLICABLE WHEN THE GYROSCOPE IS
  // DISABLED! Otherwise accel sample rate = gyro sample rate.
  // accel sample rate can be 1-6
  // 1 = 10 Hz    4 = 238 Hz
  // 2 = 50 Hz    5 = 476 Hz
  // 3 = 119 Hz   6 = 952 Hz
  imu.settings.accel.sampleRate = 1; // Set accel to 238Hz.
  // [bandwidth] sets the anti-aliasing filter bandwidth.
  // Accel cutoff freqeuncy can be any value between -1 - 3. 
  // -1 = bandwidth determined by sample rate
  // 0 = 408 Hz   2 = 105 Hz
  // 1 = 211 Hz   3 = 50 Hz
  imu.s QAettings.accel.bandwidth = -1; // BW is dependent on sample rate
  // [highResEnable] enables or disables high resolution 
  // mode for the acclerometer.
  imu.settings.accel.highResEnable = false; // Ensable HR
  // [highResBandwidth] sets the LP cutoff frequency of
  // the accelerometer if it's in high-res mode.
  // can be any value between 0-3
  // LP cutoff is set to a factor of sample rate
  // 0 = ODR/50    2 = ODR/9
  // 1 = ODR/100   3 = ODR/400
  imu.settings.accel.highResBandwidth = 0;  
}
*/
/******** Configure INT1 for Acceleration Interrupt **********/

/*
void configureLSM9DS1Interrupts()
{
  // 1. Configure accelerometer interrupt generator:
  //   - XHIE_XL: x-axis high event
  //     More axis events can be or'd together
  //   - false: OR interrupts
  //imu.configAccelInt(XHIE_XL||YHIE_XL, false);
  imu.configAccelInt(XHIE_XL, false);
  
  // 2. Configure X-axis accelerometer threshold:
  //   - 20: Threshold (raw value from accel: 0-255)
  //     Multiply this value by 128 to get threshold value.
  //     (105 = 13,440 raw accel value)
  //   - X_AXIS: Write to X-axis threshold
  //   - 10: duration until triggering interrupt (based on ODR)
  //   - true: wait (wait [duration] before interrupt goes low)
  imu.configAccelThs(105, X_AXIS, 1000, false);

/*
 * one step at a time
  // 3. Configure Y-axis accelerometer threshold:
  //   - 20: Threshold (raw value from accel: 0-255)
  //     Multiply this value by 128 to get threshold value.
  //     (20 = 2600 raw accel value)
  //   - Y_AXIS: Write to Y-axis threshold
  //   - 10: duration (based on ODR)
  //   - true: wait (wait [duration] before interrupt goes low)
  imu.configAccelThs(40, Y_AXIS, 10, true);
*
 
  // 4. Configure INT1 - assign it to accel interrupt
  //   - XG_INT1: Says we're configuring INT1
  //   - INT1_IG_XL: Sets interrupt source to the accel interrupt generator
  //   - INT_ACTIVE_LOW: Sets interrupt to active low.
  //         (Can otherwise be set to INT_ACTIVE_HIGH.)
  //   - INT_PUSH_PULL: Sets interrupt to a push-pull
  //         (Can otherwise be set to INT_OPEN_DRAIN.)
  imu.configInt(XG_INT1, INT_IG_XL, INT_ACTIVE_LOW, INT_PUSH_PULL);  
}
*/
void IRCollisions()
{
    /* Calculate Hit Count*/
    frontCm = frontSensor.distance();
    backCm  = backSensor.distance();
    leftCm  = leftSensor.distance();
    rightCm = rightSensor.distance();
    // for loop to iterate through distArray
    for (i=0; i<4; i++) {
      //compare each distance to the minimum distance that can be detected, if below min cm, hitcount++
      ArrayValue = distArray[i];
      if (ArrayValue <= minDist) {
          irCollisions++;
      }
     }
}
/******Accelerometer Printing Function ******/
/*testing
void printAccel() //EDIT THIS ****** WILL CAUSE ERROR, USE MAIN LOOP
{
  // Now we can use the ax, ay, and az variables as we please.
  // Either print them as raw ADC values, or calculated in g's.
  Serial.print("Acc: ");
#ifdef PRINT_CALCULATED
  // If you want to print calculated values, you can use the
  // calcAccel helper function to convert a raw ADC value to
  // g's. Give the function the value that you want to convert.
  Serial.print(imu.calcAccel(imu.ax), 3);
  Serial.print(", ");
  Serial.print(imu.calcAccel(imu.ay), 3);
  Serial.print(", ");
  Serial.print(imu.calcAccel(imu.az), 3);
  Serial.print(" g");
  Serial.print("\t\t");
  Serial.print(analogRead(INT1_PIN_THS));
  Serial.print("\t\t");
  Serial.print(digitalRead(INT1_PIN_THS));
  Serial.print("\n");

  Serial.print("Raw: ");
  Serial.print(imu.ax);
  Serial.print(", ");
  Serial.print(imu.ay);
  Serial.print(", ");
  Serial.println(imu.az);
#endif
}*/

